#!/usr/bin/perl -w
#
# Joe Rothstein; Dec. 13, 2006
#
# NOTES:  Input datafile.dat must be in either LINKMAP OR MLINK
# format.  The trait needs to be the first locus, and the marker
# loci should be in physical order (i.e. the third line should
# be "1 2 3 ... N", where N is one more than the number of marker
# loci.
# Only binary traits for now. No liability classes allowed.

use strict;
use IO::File;

my %file_name = (ped => "./pedfile.dat",
	par => "./datafile.dat",
	alleles => "./alleles");
my %output_file_name = (lmm_ped => "./lmm_ped",
	lmm_par => "./lmm_par",
	lmm_mrks => "./lmm_mrks");
my %file_handle;
for (sort keys %file_name) {
	$file_handle{$_} = new IO::File;
	$file_handle{$_}->open("<$file_name{$_}") or die "can't open $file_name{$_}\n";
}
for (sort keys %output_file_name) {
	$file_handle{$_} = new IO::File;
	$file_handle{$_}->open(">$output_file_name{$_}") or die "can't open $file_name{$_}\n";
}

my $Nmrks;
my @penetrances;
my @trait_freqs;
my @Nalleles;
my @mrk_freqs;
my @recombs;
my @id_order;
my %dad;
my %mom;
my %sex;
my %trait;
my %mrks;
my @allele_key;

{
	# read the parameters from the linkage "datafile"
	my $handle = $file_handle{par};  # there's got to be a better way to do this!
	while (<$handle>) {
		chomp;
		my @F = split;
		if ($. == 1) {
			$Nmrks = $F[0] - 1;
			next;
		}
		if ($. == 3) {
			if (@F != $Nmrks + 1) {
				die "lines 1 and 3 don't match";
			}
			for (0..$#F) {
				if ($F[$_] != $_+1) {
					die "loci not in proper order";
				}
			}
			next;
		}
		if ($. == 4) {
			if ($F[0] ne "1" or $F[1] ne "2") {
				die "can't handle this trait";
			}
			next;
		}
		if ($. == 5) {
			@trait_freqs = @F[0..1];
			next;
		}
		if ($. == 6 && $F[0] ne "1") {
			die "can't handle liability classes";
			next;
		}
		if ($. == 7) {
			@penetrances = @F[0..2];
			next;
		}
		if ($. > 7  &&  $. <= 7 + 2*$Nmrks) {
			# should be reading in marker allele freqs
			if ( ($. - 8) % 2 == 0) {
				my $mrk = ($. - 6) / 2;
				die "unknown marker type" if $F[0] != 3;
				$Nalleles[$mrk-1] = $F[1];
			}
			if ( ($. - 8) % 2 == 1) {
				my $mrk = ($. - 7) / 2;
				$mrk_freqs[$mrk-1] = [@F[0..($Nalleles[$mrk-1]-1)]];
			}
		}
		if ($. == 7 + $Nmrks*2 + 2) {
			die "wrong number of recombination fractions on line $." if (@F != $Nmrks);
			@recombs = @F[1..$#F];
		}
		
	}
}

{
	# read the allele conversion file
	my $handle = $file_handle{alleles};
	while (<$handle>) {
		chomp;
		my @F = split;
		$allele_key[$.-1]{0} = 0; # leave the missing data alone
		$allele_key[$.-1]{$F[$_]} = $_+1 for (0..$#F);
	}
}

{
	# read the pedigree file
	my $handle = $file_handle{ped};
	while (<$handle>) {
		chomp;
		my @F = split;
		die "wrong number of fields" if (@F != 6 + 2*$Nmrks);
		my $uniq_id = "$F[0]_$F[1]";
		$id_order[$.-1] = $uniq_id;
		$dad{$uniq_id} = ($F[2] eq "0") ? $F[2] : "$F[0]_$F[2]";
		$mom{$uniq_id} = ($F[3] eq "0") ? $F[3] : "$F[0]_$F[3]";
		$sex{$uniq_id} = $F[4];
		$trait{$uniq_id} = $F[5];
		$mrks{$uniq_id} = [@F[6..$#F]];
	}
}

{
	my $old_handle = select $file_handle{lmm_par};
	our $\ = "\n";

	(my $output = <<EOF) =~ s/^[ \t]*//gm;
		input pedigree file '$output_file_name{lmm_ped}'
		# in case the pedigree needs to be put in chronological order\noutput overwrite pedigree file '$output_file_name{lmm_ped}-chron'
		input marker data file '$output_file_name{lmm_mrks}'
		input seed file 'seeds'
		output seed file 'seeds'
		
		select all markers traits 1
		
		set MC iterations ____
		# set burn-in iterations ____ # default is 1/10 of the number of "MC iterations"
		# use ____ sequential imputation realizations for setup # default is 1/10 of the number of "MC iterations"
		# set L-sampler probability ____ # default is 0.2
		
		# trait model parameters
		set trait data discrete
		set trait 1 freqs @trait_freqs
		set incomplete penetrances @penetrances

		# positions of lod score computations
		map trait 1 all intervals proportions 0.2 0.5 0.8
		map trait 1 external distances 5 10 20 30 40 50

EOF

	print $output;
	{
		my @map = (0);
		push(@map, $map[$#map] + &haldane($_)) for (@recombs);
		print "map marker positions ", join(" ", map {sprintf "%.4f", $_} @map);
	}
	print "# map marker recomb fracts @recombs";
	print "set marker $_ freqs @{$mrk_freqs[$_-1]}" for (1..$Nmrks);
	for (1..$#allele_key+1) {
		my $key_ref = $allele_key[$_-1];
		printf "# marker $_ allele conversion (old => new): ";
		my @stuff = map { "$_ => " . ${$key_ref}{$_} } (sort {$a <=> $b} keys %{$key_ref});
		# my @stuff = map { "$_ => $_" } (sort {$a <=> $b} keys %{$key_ref});
		print join ", ", @stuff;
	}

	select $old_handle;
}

{
	# create pedigree file
	my $old_handle = select $file_handle{lmm_ped};
	our $\ = "\n";

	print "input pedigree size ", scalar(@id_order);
	print "input pedigree record names 3 integers 2";
	print "******";

	printf "%10s %10s %10s  %s  %s\n", $_, $dad{$_}, $mom{$_}, $sex{$_}, $trait{$_} for (@id_order);

	select $old_handle;
}

{
	# create marker data file
	my $old_handle = select $file_handle{lmm_mrks};
	our $\ = "\n";

	for (1..$#allele_key+1) {
		my $key_ref = $allele_key[$_-1];
		printf "# marker $_ allele conversion (old => new): ";
		my @stuff = map { "$_ => " . ${$key_ref}{$_} } (sort {$a <=> $b} keys %{$key_ref});
		# my @stuff = map { "$_ => $_" } (sort {$a <=> $b} keys %{$key_ref});
		print join ", ", @stuff;
	}

	print "set markers $Nmrks data";
	# print "$_ @{$mrks{$_}}" for (@id_order);
	for my $id (@id_order) {
		my @OF = ($id);
		for my $mrk (1..$Nmrks) {
			my @genotype = @{$mrks{$id}}[2*$mrk-2,2*$mrk-1];
			@genotype = map {$allele_key[$mrk-1]{$_}} @genotype;
			push (@OF, @genotype);
		}
		print "@OF";
	}


	select $old_handle;
}
	
sub haldane {
	100*(-0.5)*log(1-2*$_[0]);
}
